import atexit
import logging
import os
import importlib.util


from colorama import Fore
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as GRPCExporter,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider, SpanProcessor
from opentelemetry.sdk.trace.export import (
    SpanExporter,
    SimpleSpanProcessor,
    BatchSpanProcessor,
)
from opentelemetry.trace import get_tracer_provider, ProxyTracerProvider
from opentelemetry.context import get_value, attach, set_value

from elixir import Telemetry
from elixir.instruments import Instruments
from elixir.tracing.semconv import ElixirContextValues, SpanAttributes
from elixir.utils.ipython import is_notebook
from typing import Dict, Optional, Set

TRACER_NAME = "elixir.tracer"
EXCLUDED_URLS = """
    iam.cloud.ibm.com,
    dataplatform.cloud.ibm.com,
    ml.cloud.ibm.com,
    api.openai.com,
    openai.azure.com,
    api.anthropic.com,
    api.cohere.ai,
    pinecone.io,
    posthog.com,
    sentry.io,
    bedrock-runtime,
    googleapis.com,
    githubusercontent.com,
    openaipublic.blob.core.windows.net"""


class TracerWrapper(object):
    resource_attributes: dict = {}
    endpoint: str = None
    headers: Dict[str, str] = {}

    def __new__(
        cls,
        disable_batch=False,
        should_enrich_metrics: bool = True,
        instruments: Optional[Set[Instruments]] = None,
        exporter: SpanExporter = None,
    ) -> "TracerWrapper":
        if not hasattr(cls, "instance"):
            obj = cls.instance = super(TracerWrapper, cls).__new__(cls)
            if not TracerWrapper.endpoint:
                return obj

            obj.__resource = Resource(attributes=TracerWrapper.resource_attributes)
            obj.__tracer_provider: TracerProvider = init_tracer_provider(
                resource=obj.__resource
            )

            Telemetry().capture(
                "tracer:init",
                {
                    "exporter": "custom" if exporter else TracerWrapper.endpoint,
                    "processor": "simple" if disable_batch else "batch",
                },
            )

            obj.__spans_exporter: SpanExporter = (
                exporter
                if exporter
                else init_spans_exporter(TracerWrapper.endpoint, TracerWrapper.headers)
            )
            if disable_batch or is_notebook():
                obj.__spans_processor: SpanProcessor = SimpleSpanProcessor(
                    obj.__spans_exporter
                )
            else:
                obj.__spans_processor: SpanProcessor = BatchSpanProcessor(
                    obj.__spans_exporter
                )
            obj.__spans_processor_original_on_start = None

            obj.__spans_processor.on_start = obj._span_processor_on_start
            obj.__tracer_provider.add_span_processor(obj.__spans_processor)

            instrument_set = False
            if instruments is None:
                init_instrumentations(should_enrich_metrics)
                instrument_set = True
            else:
                for instrument in instruments:
                    if instrument == Instruments.OPENAI:
                        if not init_openai_instrumentor(should_enrich_metrics):
                            print(Fore.RED + "Warning: OpenAI library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.ANTHROPIC:
                        if not init_anthropic_instrumentor(should_enrich_metrics):
                            print(
                                Fore.RED + "Warning: Anthropic library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.COHERE:
                        if not init_cohere_instrumentor():
                            print(Fore.RED + "Warning: Cohere library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.PINECONE:
                        if not init_pinecone_instrumentor():
                            print(
                                Fore.RED + "Warning: Pinecone library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.CHROMA:
                        if not init_chroma_instrumentor():
                            print(Fore.RED + "Warning: Chroma library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.LANGCHAIN:
                        if not init_langchain_instrumentor():
                            print(
                                Fore.RED + "Warning: LangChain library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.MISTRAL:
                        if not init_mistralai_instrumentor():
                            print(
                                Fore.RED + "Warning: MistralAI library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.OLLAMA:
                        if not init_ollama_instrumentor():
                            print(Fore.RED + "Warning: Ollama library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.LLAMA_INDEX:
                        if not init_llama_index_instrumentor():
                            print(
                                Fore.RED + "Warning: LlamaIndex library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.MILVUS:
                        if not init_milvus_instrumentor():
                            print(Fore.RED + "Warning: Milvus library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.TRANSFORMERS:
                        if not init_transformers_instrumentor():
                            print(
                                Fore.RED
                                + "Warning: Transformers library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.BEDROCK:
                        if not init_bedrock_instrumentor(should_enrich_metrics):
                            print(Fore.RED + "Warning: Bedrock library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.REPLICATE:
                        if not init_replicate_instrumentor():
                            print(
                                Fore.RED + "Warning: Replicate library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.VERTEXAI:
                        if not init_vertexai_instrumentor():
                            print(
                                Fore.RED + "Warning: Vertex AI library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.WATSONX:
                        if not init_watsonx_instrumentor():
                            print(Fore.RED + "Warning: Watsonx library does not exist.")
                            print(Fore.RESET)
                        else:
                            instrument_set = True
                    elif instrument == Instruments.WEAVIATE:
                        if not init_weaviate_instrumentor():
                            print(
                                Fore.RED + "Warning: Weaviate library does not exist."
                            )
                            print(Fore.RESET)
                        else:
                            instrument_set = True

                    else:
                        print(
                            Fore.RED
                            + "Warning: "
                            + instrument
                            + " instrumentation does not exist."
                        )
                        print(
                            "Usage:\n"
                            + "from elixir.instruments import Instruments\n"
                            + 'Elixir.init(app_name="...", instruments=set([Instruments.OPENAI]))'
                        )
                        print(Fore.RESET)

            if not instrument_set:
                print(
                    Fore.RED + "Warning: No valid instruments set. Remove 'instrument' "
                    "argument to use all instruments, or set a valid instrument."
                )
                print(Fore.RESET)

            # Force flushes for debug environments (e.g. local development)
            atexit.register(obj.exit_handler)

        return cls.instance

    def exit_handler(self):
        self.flush()

    def _span_processor_on_start(self, span, parent_context):
        entity_name = get_value(ElixirContextValues.ENTITY_NAME)
        if entity_name is not None:
            span.set_attribute(SpanAttributes.ELIXIR_ENTITY_NAME, entity_name)

        association_properties = get_value(ElixirContextValues.ASSOCIATION_PROPERTIES)
        if association_properties is not None:
            for key, value in association_properties.items():
                span.set_attribute(
                    f"{SpanAttributes.ELIXIR_ASSOCIATION_PROPERTIES}.{key}", value
                )

        # Call original on_start method if it exists in custom processor
        if self.__spans_processor_original_on_start:
            self.__spans_processor_original_on_start(span, parent_context)

    @staticmethod
    def set_static_params(
        resource_attributes: dict,
        endpoint: str,
        headers: Dict[str, str],
    ) -> None:
        TracerWrapper.resource_attributes = resource_attributes
        TracerWrapper.endpoint = endpoint
        TracerWrapper.headers = headers

    @classmethod
    def verify_initialized(cls) -> bool:
        if hasattr(cls, "instance"):
            return True

        if (os.getenv("ELIXIR_SUPPRESS_WARNINGS") or "false").lower() == "true":
            return False

        print(
            Fore.RED
            + "Warning: Elixir not initialized, make sure you call Elixir.init()"
        )
        print(Fore.RESET)
        return False

    def flush(self):
        self.__spans_processor.force_flush()

    def get_tracer(self):
        return self.__tracer_provider.get_tracer(TRACER_NAME)


def set_association_properties(properties: dict) -> None:
    attach(set_value(ElixirContextValues.ASSOCIATION_PROPERTIES, properties))


def set_entity_name(entity_name: str) -> None:
    attach(set_value(ElixirContextValues.ENTITY_NAME, entity_name))


def get_chained_entity_name(entity_name: str) -> str:
    parent = get_value(ElixirContextValues.ENTITY_NAME)
    if parent is None:
        return entity_name
    else:
        return f"{parent}.{entity_name}"


def init_spans_exporter(api_endpoint: str, headers: Dict[str, str]) -> SpanExporter:
    if "http" in api_endpoint.lower():
        return HTTPExporter(endpoint=f"{api_endpoint}/traces", headers=headers)
    else:
        return GRPCExporter(endpoint=f"{api_endpoint}", headers=headers)


def init_tracer_provider(resource: Resource) -> TracerProvider:
    provider: TracerProvider = None
    default_provider: TracerProvider = get_tracer_provider()

    if isinstance(default_provider, ProxyTracerProvider):
        provider = TracerProvider(resource=resource)
        trace.set_tracer_provider(provider)
    elif not hasattr(default_provider, "add_span_processor"):
        logging.error(
            "Cannot add span processor to the default provider since it doesn't support it"
        )
        return
    else:
        provider = default_provider

    return provider


def init_instrumentations(should_enrich_metrics: bool):
    init_openai_instrumentor(should_enrich_metrics)
    init_anthropic_instrumentor(should_enrich_metrics)
    init_cohere_instrumentor()
    init_pinecone_instrumentor()
    init_qdrant_instrumentor()
    init_chroma_instrumentor()
    init_haystack_instrumentor()
    init_langchain_instrumentor()
    init_mistralai_instrumentor()
    init_ollama_instrumentor()
    init_llama_index_instrumentor()
    init_milvus_instrumentor()
    init_transformers_instrumentor()
    init_bedrock_instrumentor(should_enrich_metrics)
    init_replicate_instrumentor()
    init_vertexai_instrumentor()
    init_watsonx_instrumentor()
    init_weaviate_instrumentor()


def init_openai_instrumentor(should_enrich_metrics: bool):
    try:
        if importlib.util.find_spec("openai") is not None:
            Telemetry().capture("instrumentation:openai:init")
            from opentelemetry.instrumentation.openai import OpenAIInstrumentor

            instrumentor = OpenAIInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
                enrich_assistant=should_enrich_metrics,
                enrich_token_usage=should_enrich_metrics,
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing OpenAI instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_anthropic_instrumentor(should_enrich_metrics: bool):
    try:
        if importlib.util.find_spec("anthropic") is not None:
            Telemetry().capture("instrumentation:anthropic:init")
            from opentelemetry.instrumentation.anthropic import AnthropicInstrumentor

            instrumentor = AnthropicInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
                enrich_token_usage=should_enrich_metrics,
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Anthropic instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_cohere_instrumentor():
    try:
        if importlib.util.find_spec("cohere") is not None:
            Telemetry().capture("instrumentation:cohere:init")
            from opentelemetry.instrumentation.cohere import CohereInstrumentor

            instrumentor = CohereInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Cohere instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_pinecone_instrumentor():
    try:
        if importlib.util.find_spec("pinecone") is not None:
            Telemetry().capture("instrumentation:pinecone:init")
            from opentelemetry.instrumentation.pinecone import PineconeInstrumentor

            instrumentor = PineconeInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Pinecone instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_qdrant_instrumentor():
    try:
        if importlib.util.find_spec("qdrant_client") is not None:
            Telemetry().capture("instrumentation:qdrant:init")
            from opentelemetry.instrumentation.qdrant import QdrantInstrumentor

            instrumentor = QdrantInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
    except Exception as e:
        logging.error(f"Error initializing Qdrant instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_chroma_instrumentor():
    try:
        if importlib.util.find_spec("chromadb") is not None:
            Telemetry().capture("instrumentation:chromadb:init")
            from opentelemetry.instrumentation.chromadb import ChromaInstrumentor

            instrumentor = ChromaInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Chroma instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_haystack_instrumentor():
    try:
        if importlib.util.find_spec("haystack") is not None:
            Telemetry().capture("instrumentation:haystack:init")
            from opentelemetry.instrumentation.haystack import HaystackInstrumentor

            instrumentor = HaystackInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Haystack instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_langchain_instrumentor():
    try:
        if importlib.util.find_spec("langchain") is not None:
            Telemetry().capture("instrumentation:langchain:init")
            from opentelemetry.instrumentation.langchain import LangchainInstrumentor

            instrumentor = LangchainInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing LangChain instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_mistralai_instrumentor():
    try:
        if importlib.util.find_spec("mistralai") is not None:
            Telemetry().capture("instrumentation:mistralai:init")
            from opentelemetry.instrumentation.mistralai import MistralAiInstrumentor

            instrumentor = MistralAiInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing MistralAI instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_ollama_instrumentor():
    try:
        if importlib.util.find_spec("ollama") is not None:
            Telemetry().capture("instrumentation:ollama:init")
            from opentelemetry.instrumentation.ollama import OllamaInstrumentor

            instrumentor = OllamaInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Ollama instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_transformers_instrumentor():
    try:
        if importlib.util.find_spec("transformers") is not None:
            Telemetry().capture("instrumentation:transformers:init")
            from opentelemetry.instrumentation.transformers import (
                TransformersInstrumentor,
            )

            instrumentor = TransformersInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Transformers instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_llama_index_instrumentor():
    try:
        if importlib.util.find_spec("llama_index") is not None:
            Telemetry().capture("instrumentation:llamaindex:init")
            from opentelemetry.instrumentation.llamaindex import LlamaIndexInstrumentor

            instrumentor = LlamaIndexInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing LlamaIndex instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_milvus_instrumentor():
    try:
        if importlib.util.find_spec("pymilvus") is not None:
            Telemetry().capture("instrumentation:milvus:init")
            from opentelemetry.instrumentation.milvus import MilvusInstrumentor

            instrumentor = MilvusInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Milvus instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_bedrock_instrumentor(should_enrich_metrics: bool):
    try:
        if importlib.util.find_spec("boto3") is not None:
            from opentelemetry.instrumentation.bedrock import BedrockInstrumentor

            instrumentor = BedrockInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
                enrich_token_usage=should_enrich_metrics,
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Bedrock instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_replicate_instrumentor():
    try:
        if importlib.util.find_spec("replicate") is not None:
            Telemetry().capture("instrumentation:replicate:init")
            from opentelemetry.instrumentation.replicate import ReplicateInstrumentor

            instrumentor = ReplicateInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Replicate instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_vertexai_instrumentor():
    try:
        if importlib.util.find_spec("vertexai") is not None:
            Telemetry().capture("instrumentation:vertexai:init")
            from opentelemetry.instrumentation.vertexai import VertexAIInstrumentor

            instrumentor = VertexAIInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Vertex AI instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_watsonx_instrumentor():
    try:
        if importlib.util.find_spec("ibm_watson_machine_learning") is not None:
            Telemetry().capture("instrumentation:watsonx:init")
            from opentelemetry.instrumentation.watsonx import WatsonxInstrumentor

            instrumentor = WatsonxInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Watsonx instrumentor: {e}")
        Telemetry().log_exception(e)
        return False


def init_weaviate_instrumentor():
    try:
        if importlib.util.find_spec("weaviate") is not None:
            Telemetry().capture("instrumentation:weaviate:init")
            from opentelemetry.instrumentation.weaviate import WeaviateInstrumentor

            instrumentor = WeaviateInstrumentor(
                exception_logger=lambda e: Telemetry().log_exception(e),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Weaviate instrumentor: {e}")
        Telemetry().log_exception(e)
        return False
